import pandas as pd
import plotly.express as px
from .px_util import do_prefixes


def do_barchart(settings: dict, data: pd.DataFrame, label=None):
    x_col = settings.get("x", ["label"])[0]
    y_col = settings.get("y", ["value"])[0]
    data[y_col] = data[y_col].astype(float)
    data[x_col] = data[x_col].apply(do_prefixes)
    return px.bar(data, x=x_col, y=y_col, title=label)


def do_piechart(settings: dict, data: pd.DataFrame, label=None):
    label = settings.get("label", [None])[0]
    values = settings.get("values", ["value"])[0]
    names = settings.get("names", ["label"])[0]
    data[values] = data[values].astype(float)
    data[names] = data[names].apply(do_prefixes)
    return px.pie(data, values=values, names=names, title=label)


def do_mapchart(settings: dict, data: pd.DataFrame, label=None):
    point = settings.get("point", ["geo"])[0]

    if point in data.columns:

        def safe_extract_lat(point):
            try:
                return float(point.upper().replace("POINT(", "").split(" ")[0])
            except Exception:
                return None

        def safe_extract_lon(point):
            try:
                return float(
                    point.upper().replace("POINT(", "").strip(")").split(" ")[1]
                )
            except Exception:
                return None

        data["lon"] = data[point].apply(safe_extract_lat)
        data["lat"] = data[point].apply(safe_extract_lon)

        # Remove rows where lat or lon is None
        data = data.dropna(subset=["lat", "lon"])

    if "lat" not in data.columns or "lon" not in data.columns:
        raise ValueError(
            "Data must contain 'lat' and 'lon' columns for map chart, or a 'geo' column containing Point(x y) format."
        )

    lat_col = "lat"
    lon_col = "lon"

    data[lat_col] = data[lat_col].astype(float)
    data[lon_col] = data[lon_col].astype(float)

    center_lat = float(settings.get("lat", [0])[0])
    center_lon = float(settings.get("lon", [0])[0])

    zoom = int(settings.get("zoom", [3])[0])
    return px.scatter_map(
        data,
        lat=lat_col,
        lon=lon_col,
        zoom=zoom,
        title=label,
        center={"lat": center_lat, "lon": center_lon},
    )



import httpx, logging, random, hashlib, json, time, sqlite3, os, gzip
import fizzysearch
from .config import (
    ENDPOINT,
    ENDPOINTS,
    QUERIES_DB,
    BIKIDATA_DB,
    SEMANTIC_INDEX,
    PREFIXES_SNIPPET,
    DATA_LOAD_PATHS,
    STORE_PATH,
    log,
)
import pyoxigraph as px
from .px_util import OxigraphSerialization, string_iterator


def hash_query(query: str) -> str:
    return hashlib.md5(query.encode("utf8")).hexdigest()


def cached_query(query: str, endpoint: str = None):
    # Only use the endpoint if specified
    if endpoint:
        theq = sqlite3.connect(QUERIES_DB).execute(
            "SELECT timestamp, result, duration FROM queries WHERE queryhash = ? and endpoint = ? and not result is null ORDER BY timestamp DESC LIMIT 1",
            (hash_query(query), endpoint),
        )
    else:
        theq = sqlite3.connect(QUERIES_DB).execute(
            "SELECT timestamp, result, duration FROM queries WHERE queryhash = ? and not result is null ORDER BY timestamp DESC LIMIT 1",
            (hash_query(query),),
        )

    for timestamp, result, duration in theq:
        result = json.loads(result)
        result["timestamp"] = timestamp
        result["duration"] = duration
        result["cached"] = True
        return result


def do_query(query: str) -> dict:
    to_use = ENDPOINT

    try:
        rewritten = fizzysearch.rewrite(
            query,
            {
                "https://fizzysearch.ise.fiz-karlsruhe.de/fts": fizzysearch.use_fts(),
                "fizzy:fts": fizzysearch.use_fts(),
                "fizzy:ftsStats": fizzysearch.use_fts_stats(),
            },
        )
    except Exception as e:
        log.exception(f"Problem with fizzysearch: {e}")
        return {"error": f"Fizzysearch rewriting error: {e}"}

    shmarql_settings = {}
    for comment in rewritten["comments"]:
        log.debug(f"fizzysearch SPARQL Comment: {comment}")
        if comment.find("shmarql-engine:") > -1:
            to_use = ENDPOINTS.get(comment.split(" ")[-1])
        if comment.startswith("shmarql-"):
            comment_value = [x.strip(" ") for x in comment[8:].split(":")]
            if len(comment_value) > 1:
                shmarql_settings.setdefault(comment_value[0], []).append(
                    ":".join(comment_value[1:])
                )

    query = rewritten.get("rewritten", query)

    if not to_use:
        if len(ENDPOINTS) > 0:
            to_use = random.choice(list(ENDPOINTS.values()))
        elif len(GRAPH) > 0:
            to_use = "__local__"
        else:
            return {"error": "No endpoint found"}

    if not "nocache" in shmarql_settings:
        cached_query_result = cached_query(query)
        if cached_query_result:
            return cached_query_result

    time_start = time.time()
    result = {}
    if to_use == "__local__":
        try:
            qquery = PREFIXES_SNIPPET + "\n" + query
            r = GRAPH.query(qquery, use_default_graph_as_union=True)
            result = OxigraphSerialization(r).json()
        except Exception as e:
            return {"error": str(e)}
    else:
        if rewritten.get("query_type") == "construct":
            accept_header = "text/turtle"
        else:
            accept_header = "application/sparql-results+json"
        headers = {
            "Accept": accept_header,
            "User-Agent": "SHMARQL/2024 (https://shmarql.com/ ep@epoz.org)",
        }

        data = {
            "query": PREFIXES_SNIPPET + "\n" + query,
        }
        try:
            r = httpx.post(to_use, data=data, headers=headers, timeout=180)
            if r.status_code == 200:
                try:
                    result = r.json()
                except json.JSONDecodeError:
                    result = {"data": r.content.decode("utf8")}
            elif r.status_code == 500:
                return {"error": r.text}
        except:
            log.exception(f"Problem with {to_use}")
            return {"error": "Exception raised querying endpoint"}

    time_end = time.time()
    duration = time_end - time_start

    # get the endpoint name from the to_use URL
    endpoint_name = "default"
    for k, v in ENDPOINTS.items():
        if v == to_use:
            endpoint_name = k

    if result:
        result["duration"] = duration
        result["endpoint_name"] = endpoint_name
        result["endpoint"] = to_use
        result["shmarql_settings"] = shmarql_settings

        thequerydb = sqlite3.connect(QUERIES_DB)
        thequerydb.execute(
            "INSERT INTO queries (queryhash, query, timestamp, endpoint, result, duration) VALUES (?, ?, datetime(), ?, ?, ?)",
            (hash_query(query), query, to_use, json.dumps(result), duration),
        )
        thequerydb.commit()

        return result
    else:
        return {"error": r.text, "status": r.status_code}


def initialize_graph(data_load_paths: list, store_path: str = None) -> px.Store:
    log.debug(f"Initialize graph with configs: {data_load_paths} and {store_path}")
    store_primary = True
    if store_path:
        log.debug(f"Opening store from {store_path}")
        if data_load_paths:
            # If there are multiple workers trying to load at the same time,
            # contention for the lock will happen.
            # Do a short wait to stagger start times and let one win, the rest will lock and open read_only
            time.sleep(random.random() / 2)
            try:
                GRAPH = px.Store(store_path)
                log.debug("This process won the loading contention")
            except OSError:
                log.debug("Secondary, opening store read-only")
                GRAPH = px.Store.secondary(store_path)
                store_primary = False
        else:
            log.debug("Opening store read-only")
            GRAPH = px.Store.read_only(store_path)
    else:
        GRAPH = px.Store()

    if len(GRAPH) < 1 and data_load_paths and store_primary:
        for data_load_path in data_load_paths:
            if data_load_path.startswith("http://") or data_load_path.startswith(
                "https://"
            ):
                log.debug(f"Downloading {data_load_path}")
                # Try downloading this file and parsing it as a string
                start_download = time.time()
                r = httpx.get(data_load_path, follow_redirects=True, timeout=180)
                if r.status_code == 200:
                    log.debug(
                        f"Downloading {data_load_path} took {int(time.time() - start_download)} seconds"
                    )
                    d = r.content
                    # Try and guess content type from extention, default is turtle
                    # if .rdf or .nt use on of those
                    if (
                        data_load_path.endswith(".rdf")
                        or data_load_path.endswith(".xml")
                        or data_load_path.endswith(".owl")
                    ):
                        GRAPH.bulk_load(r.content, "application/rdf+xml")
                    elif data_load_path.endswith(".nt") or data_load_path.endswith(
                        ".nt.gz"
                    ):
                        GRAPH.bulk_load(r.content, "application/n-triples")
                    else:
                        GRAPH.bulk_load(r.content, "text/turtle")
            else:
                if load_file_to_graph(GRAPH, data_load_path):
                    continue
                for dirpath, _, filenames in os.walk(data_load_path):
                    for filename in filenames:
                        try:
                            filepath = os.path.join(dirpath, filename)
                            load_file_to_graph(GRAPH, filepath)
                        except SyntaxError:
                            log.error(f"Failed to parse {filepath}")

    log.debug(f"Graph haz {len(GRAPH)} triples")

    if store_primary and BIKIDATA_DB and not os.path.exists(BIKIDATA_DB):
        r = fizzysearch.build_from_iterator(string_iterator(GRAPH))
        log.debug(
            f"NEW bikidata built at {BIKIDATA_DB} with {r.get('count', '?')} literals"
        )

    return GRAPH


def load_file_to_graph(graph: px.Store, filepath: str) -> bool:
    was_file = False
    try:
        filename = os.path.basename(filepath)
        if filename.endswith(".gz"):
            filepath = gzip.open(filepath)
            filename = filename[:-3]
        else:
            filepath = open(filepath, "rb")
        if filename.lower().endswith(".ttl"):
            log.debug(f"Parsing {filepath}")
            graph.bulk_load(filepath, "text/turtle")
            was_file = True
        elif filename.lower().endswith(".nt"):
            log.debug(f"Parsing {filepath}")
            graph.bulk_load(filepath, "application/n-triples")
            was_file = True
    except Exception as e:
        log.debug(f"{filepath} exception {e}")
    return was_file


if not (ENDPOINT or len(ENDPOINTS) > 0):
    log.debug("No ENDPOINT or ENDPOINTS defined, using local graph")
    GRAPH = initialize_graph(DATA_LOAD_PATHS, STORE_PATH)



import csv, io, string, json, os, yaml
from urllib.parse import quote
from fasthtml.common import *
import pyoxigraph as px
from .px_util import OxigraphSerialization, results_to_xml
from .config import (
    PREFIXES_SNIPPET,
    MOUNT,
    SPARQL_QUERY_UI,
    SITEDOCS_PATH,
    SCHPIEL_PATH,
    SITE_URI,
    log,
)
from plotly.offline._plotlyjs_version import __plotlyjs_version__ as plotlyjs_version
import asyncio
from typing import List, Callable, Dict, Any
from monsterui.all import Theme
from .layout import markdown_container, base

from .qry import do_query, hash_query
from .fragments import rt as fragments_rt
from .fragments import build_sparql_ui
from .am import ar as am_rt

hdrs = [
    Script(src=f"https://cdn.plot.ly/plotly-{plotlyjs_version}.min.js"),
]
hdrs.extend(Theme.blue.headers())


app = FastHTML(
    pico=False,
    hdrs=hdrs,
)

fragments_rt.to_app(app)
am_rt.to_app(app)


@app.on_event("startup")
async def startup_event():
    log.debug(f"App started up")


@app.get("/favicon.ico")
def favicon():
    return FileResponse(f"static/favicon.ico")


@app.get("/shmarql/static/{fname:path}")
@app.get(MOUNT + "shmarql/static/{fname:path}")
@app.get(MOUNT + "static/{fname:path}")
def shmarql_get_static(fname: str):
    return FileResponse(f"static/{fname}")


def make_literal_query(some_literal: dict, encode=True, limit=999):
    txt = some_literal["value"].replace("\n", " ")
    txt = txt.translate(str.maketrans("", "", string.punctuation))
    txt = [x for x in txt.split(" ") if len(x) > 1][:10]
    txt = " ".join(txt).strip(" ")

    Q = f"""select ?s ?p ?o where {{ 
        ?s ?p ?o .
        ?s fizzy:fts "{txt}" . }} limit {limit}"""
    if encode:
        return quote(Q)
    else:
        return Q


def json_results_to_csv(results: dict):
    output = io.StringIO()
    writer = csv.writer(output)

    # Write header
    writer.writerow([var for var in results.get("head", {}).get("vars", [])])
    for row in results.get("results", {}).get("bindings", []):
        writer.writerow(
            [
                row.get(var, {"value": ""})["value"]
                for var in results.get("head", {}).get("vars", [])
            ]
        )
    return output.getvalue()


def accept_header_to_format(request: Request) -> str:
    accept_header = request.headers.get("accept")
    accept_headers_incoming = []
    if accept_header:
        accept_headers_incoming = [ah.strip() for ah in accept_header.split(",")]
    accept_headers = {}
    for ah in accept_headers_incoming:
        ql = ah.split(";")
        ql_val = 1
        if len(ql) > 1:
            ah_left = ql[0]
            ql = ql[1].split("=")
            if len(ql) > 1 and ql[0] == "q":
                try:
                    ql_val = float(ql[1])
                except ValueError:
                    ql_val = 0
        else:
            ah_left = ql[0]
            ql_val = 1
        accept_headers[ah_left] = ql_val

    for ah, _ in sorted(accept_headers.items(), reverse=True, key=lambda x: x[1]):
        if ah.startswith("application/sparql-results+json"):
            return "json"
        if ah.startswith("text/turtle"):
            return "turtle"
        if ah.startswith("application/sparql-results+xml"):
            return "xml"

    return "html"


@app.options(f"{MOUNT}sparql")
def handle_options():
    headers = {
        "Access-Control-Allow-Origin": "*",
        "Access-Control-Allow-Methods": "GET, POST, OPTIONS",
        "Access-Control-Allow-Headers": "Content-Type, Authorization",
    }
    return Response(status_code=204, headers=headers)


@app.route(f"{MOUNT}sparql", methods=["GET", "POST"])
async def sparql(request: Request):
    if request.headers.get("content-type") == "application/sparql-query":
        query = await request.body()
        query = query.decode("utf-8")
    else:
        query = request.query_params.get("query")
    if not query or len(query.strip()) < 2:
        return RedirectResponse(f"{MOUNT}shmarql/", status_code=303)
    return shmarql_get(request, query)


@app.get(f"{MOUNT}shmarql")
def shmarql_redir():
    return RedirectResponse(f"{MOUNT}shmarql/", status_code=303)


@app.get(f"{MOUNT}shmarql/")
def shmarql_get(
    request: Request,
    session,
    query: str = "select * where {?s ?p ?o} limit 10",
    format: str = None,
):

    if format is None:
        format = accept_header_to_format(request)
    results = do_query(query)
    if format in ("csv", "json", "turtle", "xml"):
        if "data" in results:
            if format == "turtle":
                return Response(
                    results["data"],
                    headers={
                        "Content-Type": "text/turtle",
                        "Access-Control-Allow-Origin": "*",
                    },
                )
            try:
                # if format is not turtle, but the results are bytes, try to parse it and return as json
                tmp_store = px.Store()
                tmp_store.bulk_load(results["data"], "text/turtle")
                r = tmp_store.query("select * where {?s ?p ?o}")
                results["results"] = {"bindings": OxigraphSerialization(r).json()}
            except Exception as e:
                results = {
                    "error": f"{e} Query returned non-parsable data: {repr(results)[:500]}"
                }

        if format == "xml":
            xml_data = results_to_xml(results)
            return Response(
                xml_data,
                headers={
                    "Content-Type": "application/sparql-results+xml",
                    "Content-Disposition": f"attachment; filename={hash_query(query)}.xml",
                    "Access-Control-Allow-Origin": "*",
                },
            )

        if format == "csv":
            csv_data = json_results_to_csv(results)

            return Response(
                csv_data,
                headers={
                    "Content-Type": "text/csv",
                    "Content-Disposition": f"attachment; filename={hash_query(query)}.csv",
                    "Access-Control-Allow-Origin": "*",
                },
            )
        if format == "json":
            if "endpoint" in results:
                del results["endpoint"]
            return Response(
                json.dumps(results, indent=2),
                headers={
                    "Content-Type": "application/sparql-results+json",
                    "Access-Control-Allow-Origin": "*",
                },
            )

    if not SPARQL_QUERY_UI:
        return Div(
            "There is currently no SPARQL query form to be found here, call it from a command line via a POST request."
        )

    return base(Div(build_sparql_ui(query, results), cls="mt-4"), session=session)


from .biki import *


def entity_check(iri: str):
    q = f"SELECT * WHERE {{ <{iri}> ?p ?o }}"
    res = do_query(q)
    return len(res.get("results", {}).get("bindings", [])) > 0


@app.get(MOUNT + "{fname:path}")
@app.get("/{fname:path}")
def getter(request: Request, session, fname: str):
    log.debug(f"Getter on {repr(fname)}")
    new_name = fname
    if fname.startswith("/"):
        new_name = fname[1:]
    if fname == "" or fname.endswith("/"):
        new_name += "index.html"

    if SCHPIEL_PATH:
        path_to_try = os.path.join(SCHPIEL_PATH, new_name)
        if os.path.exists(path_to_try):
            return FileResponse(path_to_try)

    if new_name == "index.html":
        new_name = "index.md"
    path_to_try = os.path.join(SITEDOCS_PATH, new_name)
    if MOUNT:
        path_to_try = os.path.join(SITEDOCS_PATH, new_name.replace(MOUNT[1:], "", 1))
    else:
        path_to_try = os.path.join(SITEDOCS_PATH, new_name)

    nav = []
    for nav_name in ("mkdocs.yml", "mkdocs.yaml"):
        nav_path = os.path.join(SITEDOCS_PATH, nav_name)
        if os.path.exists(nav_path):
            try:
                nav = yaml.safe_load(open(nav_path, "r")).get("nav", [])
            except Exception as e:
                log.warning(f"Could not parse {nav_path}: {e}")
            break

    log.debug(f"Trying {path_to_try}")
    if os.path.exists(path_to_try):
        return markdown_container(path_to_try, nav=nav, session=session)
    if os.path.exists(path_to_try + ".md"):
        return markdown_container(path_to_try + ".md", nav=nav, session=session)

    iri = SITE_URI + fname
    log.debug(f"Entity Checking {iri}")
    if SITE_URI and entity_check(iri):
        format = accept_header_to_format(request)
        if format == "html":
            q = f"""
# shmarql-view: resource
# shmarql-editor: hide

SELECT ?p ?o ?pp ?oo WHERE {{ 
  <{iri}> ?p ?o .
  OPTIONAL {{
    ?o ?pp ?oo .
  }}
}}"""
        else:
            q = f"CONSTRUCT {{ <{iri}> ?p ?o }} WHERE {{ <{iri}> ?p ?o }}"

        return RedirectResponse(f"{MOUNT}shmarql/?query={quote(q)}")

    # The default 404 from FileResponse leaks the path, make it simpler:
    raise HTTPException(404, f"File not found {fname}")



import random
from urllib.parse import quote
from fasthtml.common import *
from monsterui.all import *
from fastlite import database
from .layout import base
from .config import ADMIN_DATABASE, log, LOGINS
from passlib.context import CryptContext
from functools import wraps

pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")


def reset_admin_password(admin_filepath: str):
    admin_database = database(admin_filepath)
    users = admin_database.t.users
    users.dataclass()
    admin_password = "".join(random.choice("01234567890abcdef") for _ in range(20))
    admin_user = users["admin"]
    admin_user.password = pwd_context.hash(admin_password)
    users.update(admin_user)
    log.info(f"*****>>> admin user password: [ {admin_password} ] <<<*****")


if LOGINS:
    admin_database = database(ADMIN_DATABASE)
    users = admin_database.t.users
    if "users" not in admin_database.t:
        users.create(
            username=str, name=str, password=str, activation_date=str, pk="username"
        )
        admin_password = "".join(random.choice("01234567890abcdef") for _ in range(20))
        users.insert(
            username="admin",
            name="Administrator",
        )
        reset_admin_password()
    users.dataclass()


ar = APIRouter()


def login_required(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        session = kwargs.get("session")
        if not session or not session.get("user"):
            return RedirectResponse(url="/login", status_code=302)
        return func(*args, **kwargs)

    return wrapper


@ar.get("/admin/users")
def admin_users(session, msg: str = None):
    if not session.get("user"):
        return RedirectResponse("/login", status_code=302)
    user = session.get("user")
    if user.get("username") != "admin":
        raise HTTPException(status_code=403, detail="Not authorized")
    content = Div(
        Div(
            Modal(
                ModalTitle("Delete User confirmation"),
                P(
                    "Are you sure you want to delete this user? This action cannot be undone.",
                    cls=TextPresets.muted_sm,
                    id="user_confirm_text",
                ),
                footer=ModalCloseButton("Close", cls=ButtonT.secondary),
                id="my-modal",
            ),
        ),
        H1("User Management", cls="text-3xl font-bold mb-6 text-center"),
        (
            Div(
                msg,
                cls="bg-red-100 border border-red-400 text-red-700 px-4 py-3 rounded mb-4",
            )
            if msg
            else None
        ),
        A("Add User", href="/admin/adduser", cls="btn btn-sm btn-primary mb-4"),
        Table(
            Thead(
                Tr(
                    Th("Username", cls="border px-4 py-2"),
                    Th("Name", cls="border px-4 py-2"),
                    Th(" ", cls="border px-4 py-2"),
                )
            ),
            Tbody(
                *[
                    Tr(
                        Td(
                            A(
                                u.username,
                                href=f"/admin/adduser?username={quote(u.username)}&name={quote(u.name)}",
                            ),
                            cls="border px-4 py-2",
                        ),
                        Td(u.name, cls="border px-4 py-2"),
                        Td(
                            (
                                Button(
                                    UkIcon(
                                        "user-round-x",
                                        height=15,
                                        width=15,
                                        title="Delete User",
                                    ),
                                    data_uk_toggle="target: #my-modal",
                                    hx_get="/admin/deleteuser?username="
                                    + quote(u.username),
                                    hx_target="#user_confirm_text",
                                    hx_swap="innerHTML",
                                )
                            ),
                            cls="border px-4 py-2",
                        ),
                    )
                    for u in users()
                    if u.username != "admin"
                ]
            ),
        ),
        cls="max-w-4xl mx-auto mt-8 p-6 bg-white rounded-lg shadow-md",
    )
    return base(content, title="User Management", session=session)


@ar.get("/admin/deleteuser")
def delete_user_get(session, username: str):
    return Div(
        "Are you sure you want to delete user ",
        B(username),
        "?",
        Button(
            "Delete",
            hx_post=f"/admin/deleteuser?username={quote(username)}",
            hx_redirect="true",
            cls="btn btn-sm btn-danger ml-2",
        ),
        cls="p-4",
    )


@ar.post("/admin/deleteuser")
def delete_user_post(session, username: str):
    if not session.get("user"):
        raise HTTPException(status_code=403, detail="Not authorized")
    user = session.get("user")
    if user.get("username") != "admin":
        raise HTTPException(status_code=403, detail="Not authorized")
    if username == "admin":
        raise HTTPException(status_code=400, detail="Cannot delete admin user")
    try:
        users.delete(username)
        msg = quote(f"User [{username}] deleted successfully")
    except Exception as e:
        msg = quote(f"Error deleting user [{username}]: {str(e)}")

    return Response(headers={"HX-Redirect": f"/admin/users?msg={msg}"})


@ar.post("/admin/adduser")
def add_user_post(session, username: str, name: str, password: str):
    if not session.get("user"):
        raise HTTPException(status_code=403, detail="Not authorized")
    user = session.get("user")
    if user.get("username") != "admin":
        raise HTTPException(status_code=403, detail="Not authorized")
    if username == "admin":
        raise HTTPException(status_code=400, detail="Cannot create another admin user")
    try:
        x_user = users[username]
        x_user.name = name
        if password:
            x_user.password = pwd_context.hash(password)
        users.update(x_user)
        msg = quote(f"User [{username}] updated successfully")
    except NotFoundError:
        if len(password) < 8:
            msg = quote("Password must be at least 8 characters long")
            return RedirectResponse(
                f"/admin/adduser?msg={msg}&username={quote(username)}&name={quote(name)}",
                status_code=302,
            )
        users.insert(
            username=username,
            name=name,
            password=pwd_context.hash(password),
        )
        msg = quote(f"User [{username}] created successfully")
    return RedirectResponse(f"/admin/users?msg={msg}", status_code=302)


@ar.get("/admin/adduser")
def add_user_get(session, username: str = "", name: str = "", msg: str = None):
    if not session.get("user"):
        return RedirectResponse("/login", status_code=302)

    heading = "Create a new User"
    btn_text = "Create"
    if username or name:
        heading = f"Modify user [{username}]"
        btn_text = "Modify"

    content = Div(
        H1(heading, cls="text-3xl font-bold mb-6 text-center"),
        (
            Div(
                msg,
                cls="bg-red-100 border border-red-400 text-red-700 px-4 py-3 rounded mb-4",
            )
            if msg
            else None
        ),
        Form(
            Div(
                Label("Username:", cls="block text-sm font-medium mb-1"),
                Input(
                    name="username",
                    value=username,
                    type="text",
                    required=True,
                    cls="w-full px-3 py-2 border border-gray-300 rounded-md focus:outline-none focus:ring-2 focus:ring-blue-500",
                ),
                cls="mb-4",
            ),
            Div(
                Label("Display name:", cls="block text-sm font-medium mb-1"),
                Input(
                    name="name",
                    value=name,
                    type="text",
                    required=True,
                    cls="w-full px-3 py-2 border border-gray-300 rounded-md focus:outline-none focus:ring-2 focus:ring-blue-500",
                ),
                cls="mb-4",
            ),
            Div(
                Label("Password:", cls="block text-sm font-medium mb-1"),
                Input(
                    name="password",
                    type="text",
                    cls="w-full px-3 py-2 border border-gray-300 rounded-md focus:outline-none focus:ring-2 focus:ring-blue-500",
                ),
                cls="mb-6",
            ),
            Button(
                btn_text,
                type="submit",
                cls="w-full bg-blue-500 text-white py-2 px-4 rounded-md hover:bg-blue-600 focus:outline-none focus:ring-2 focus:ring-blue-500",
            ),
            method="post",
            action="/admin/adduser",
            cls="space-y-4",
        ),
        cls="max-w-md mx-auto mt-8 p-6 bg-white rounded-lg shadow-md",
    )
    return base(content, title="Add User", session=session)


@ar.get("/logout")
def logout(session):
    session.clear()
    return RedirectResponse("/login", status_code=302)


@ar.post("/login")
def login(session, username: str, password: str):
    try:
        user = users.selectone(where="username=?", where_args=(username,))
        if user and pwd_context.verify(password, user.password):
            session["user"] = {"username": user.username, "name": user.name}
            return RedirectResponse("/", status_code=302)
    except Exception as e:
        log.debug(f"Login error: {str(e)}")
    return login_page(session, msg="Invalid username or password")


@ar.get("/login")
def login_page(session, msg: str | None = None):
    if session.get("user"):
        return RedirectResponse("/", status_code=302)

    content = (
        Div(
            H1("Login", cls="text-3xl font-bold mb-6 text-center"),
            (
                Div(
                    msg,
                    cls="bg-red-100 border border-red-400 text-red-700 px-4 py-3 rounded mb-4",
                )
                if msg
                else None
            ),
            Form(
                Div(
                    Label("Username:", cls="block text-sm font-medium mb-1"),
                    Input(
                        name="username",
                        type="text",
                        required=True,
                        cls="w-full px-3 py-2 border border-gray-300 rounded-md focus:outline-none focus:ring-2 focus:ring-blue-500",
                    ),
                    cls="mb-4",
                ),
                Div(
                    Label("Password:", cls="block text-sm font-medium mb-1"),
                    Input(
                        name="password",
                        type="password",
                        required=True,
                        cls="w-full px-3 py-2 border border-gray-300 rounded-md focus:outline-none focus:ring-2 focus:ring-blue-500",
                    ),
                    cls="mb-6",
                ),
                Button(
                    "Login",
                    type="submit",
                    cls="w-full bg-blue-500 text-white py-2 px-4 rounded-md hover:bg-blue-600 focus:outline-none focus:ring-2 focus:ring-blue-500",
                ),
                method="post",
                action="/login",
                cls="space-y-4",
            ),
            cls="max-w-md mx-auto mt-8 p-6 bg-white rounded-lg shadow-md",
        ),
    )

    return base(content, title="Login")



from pygments.lexers.rdf import SparqlLexer
from markdown.extensions.codehilite import CodeHiliteExtension
from markdown import Markdown
from pygments.lexers import find_lexer_class_by_name
from pygments.util import ClassNotFound

SparqlLexer.aliases.append("shmarql")

__all__ = ["ShmarqlLexer"]


class ShmarqlLexer(SparqlLexer):
    """A custom lexer for SPARQL syntax, mirroring the default SparqlLexer."""

    name = "shmarql"
    aliases = ["shmarql"]


class CustomCodeBlockExtension(CodeHiliteExtension):
    """Extend the CodeHilite extension to recognize `shmarql` blocks."""

    def extendMarkdown(self, md: Markdown):
        super().extendMarkdown(md)
        try:
            _ = find_lexer_class_by_name("shmarql")
        except ClassNotFound:
            from pygments.lexers import LEXERS

            LEXERS["ShmarqlLexer"] = (
                "shmarql.markdownplugin",
                ShmarqlLexer.name,
                tuple(ShmarqlLexer.aliases),
                ShmarqlLexer.filenames,
                ShmarqlLexer.mimetypes,
            )


def makeExtension(**kwargs):
    return CustomCodeBlockExtension(lang="shmarql")



from urllib.parse import quote
import random, json
from fasthtml.common import *
from monsterui.all import *

from .config import MOUNT, PREFIXES_SNIPPET, PREFIXES
from uuid import uuid4

from plotly.io import to_json
from .px_util import results_to_df, do_prefixes
import traceback
from .charts import do_barchart, do_piechart, do_mapchart

rt = APIRouter()

from .qry import do_query

BTN_STYLE = "bg-slate-300 hover:bg-slate-400 text-black px-2 rounded-lg shadow-xl transition duration-300 font-bold"


class HashableResult:
    """
    This class is used to make the results of a SPARQL query hashable.
    Convenience that you can then add them to a set or dict, or sort them.
    """

    def __init__(self, value):
        self.value = value

    def __hash__(self):
        return hash(self.value.get("value"))

    def __eq__(self, other):
        return isinstance(other, HashableResult) and self.value.get(
            "value"
        ) == other.value.get("value")

    def __lt__(self, other):
        return self.value.get("value") < other.value.get("value")

    def __gt__(self, other):
        return self.value.get("value") > other.value.get("value")

    def __getitem__(self, key):
        return self.value[key]

    def __setitem__(self, key, value):
        self.value[key] = value

    def __delitem__(self, key):
        del self.value[key]

    def __iter__(self):
        return iter(self.value)

    def __len__(self):
        return len(self.value)

    def __contains__(self, key):
        return key in self.value

    def __repr__(self):
        return str(self.value.get("value"))


def make_resource_query(uri: str, encode=True, limit=999):
    Q = f"""
# shmarql-view: resource
# shmarql-editor: hide

SELECT ?resource ?p ?o ?pp ?oo WHERE {{
    values ?resource {{ <{uri}> }}

  ?resource ?p ?o .
  OPTIONAL {{
    ?o ?pp ?oo .
  }}
}}  limit {limit}"""

    if encode:
        return quote(Q)
    else:
        return Q


def make_spo(uri: str, spo: str, encode=True, limit=999, extra=""):
    uri = f"<{uri}>"
    if spo not in ("s", "p", "o"):
        return f"select ?s ?p ?o where {{ ?s ?p ?o }} limit {limit}"

    if spo == "s":
        tolabel = "o"
    else:
        tolabel = "s"

    Q = f"""{extra}select ?s ?p ?o ?{tolabel}label where {{ 
  values ?{spo} {{ {uri} }} 
  ?s ?p ?o .
  optional {{
    ?{tolabel} rdfs:label ?{tolabel}label .    
  }}
  
}} limit {limit}"""

    if encode:
        return quote(Q)
    else:
        return Q


@rt.post(f"/shmarql/fragments/sparql")
@rt.post(f"{MOUNT}shmarql/fragments/sparql")
def fragments_sparql(query: str, results=None):
    if query == "":
        query = "select * where {?s ?p ?o} limit 10"
    if results is None:
        results = do_query(query)

    settings = results.get("shmarql_settings", {})

    if "data" in results:  # this was a construct query
        return Div(Pre(results["data"]))
    if "error" in results:
        return (
            Div(
                "This query did not work.",
                A(
                    "Click here to see the details of error message",
                    href="#",
                    onclick="document.getElementById('error').style.display = 'block';",
                ),
                Div(results["error"], id="error", style="display:none"),
                style="max-height: 30vh; overflow: auto;",
            ),
        )

    if "resource" in settings.get("view", []):
        return Html(fragments_resource(results, query))
    if settings.get("view", [""])[0].endswith("chart"):
        return Html(fragments_chart(query))
    else:
        return Html(build_plain_table(query, results))


def build_plain_table(query: str, results: dict):
    table_rows = []
    heads = [Th("#", style="color: #aaa; text-align: right;")]
    heads.extend(
        [
            Th(var, style="font-weight: bold")
            for var in results.get("head", {}).get("vars", [])
        ]
    )
    rownum = 0
    for row in results.get("results", {}).get("bindings", []):
        rownum += 1
        row_columns = []
        row_columns.append(
            Td(
                rownum,
                style="color: #aaa; text-align: right;",
            )
        )
        for var in results.get("head", {}).get("vars", []):
            value = row.get(var, {"value": ""})
            if value.get("type") == "uri":
                S_query = make_spo(value["value"], "s")
                P_query = make_spo(value["value"], "p")
                O_query = make_spo(value["value"], "o")

                nonce = random.randint(0, pow(2, 32))

                if var in ("s", "p", "o"):
                    var_spo = var
                else:
                    var_spo = "s"

                row_columns.append(
                    Td(
                        A(
                            do_prefixes(value["value"]),
                            href=f"{MOUNT}shmarql/?query="
                            + make_resource_query(value["value"]),
                        ),
                        Div(
                            A(
                                "S",
                                href=f"{MOUNT}shmarql/?query={S_query}",
                                style="font-size: 70%; background-color: #ddd; color: #000; padding: 3px; text-decoration: none; margin: 0",
                            ),
                            A(
                                "P",
                                href=f"{MOUNT}shmarql/?query={P_query}",
                                style="font-size: 70%; background-color: #ddd; color: #000; padding: 3px; text-decoration: none; margin: 0",
                            ),
                            A(
                                "O",
                                href=f"{MOUNT}shmarql/?query={O_query}",
                                style="font-size: 70%; background-color: #ddd; color: #000; padding: 3px; text-decoration: none; margin: 0",
                            ),
                            style="font-size: 80%; display: inline-block; margin-left: 0.5em;",
                        ),
                    )
                )
            elif value.get("type") == "bnode":
                row_columns.append(
                    Td(
                        Span(
                            value["value"], style="font-size: 80%; font-style: italic"
                        ),
                    )
                )
            else:
                lang = (
                    Span(
                        value.get("xml:lang"),
                        style="font-size: 80%; vertical-align: super;",
                    )
                    if "xml:lang" in value
                    else None
                )

                row_columns.append(Td(Span(value["value"]), lang))
        table_rows.append(Tr(*row_columns))
    cached = " (from cache) " if results.get("cached") else ""

    duration_display = (
        f"{int(results.get('duration', 0) * 1000)}ms"
        if results.get("duration", 0) < 1
        else f"{results.get('duration', 0):.3f}s"
    )

    return Div(
        P(
            f"{len(results.get('results', {}).get('bindings', []))} results in {duration_display}{cached}",
            style="font-size: 50%;",
            title="used: " + results.get("endpoint_name", ""),
        ),
        Table(Thead(Tr(*heads)), Tbody(*table_rows), data_tipe="sparql-results"),
    )


def fragments_resource(results: dict, query: str):
    """Assuming that in results is a query:
    SELECT ?p ?o ?pp ?oo WHERE {
      <some_uri> ?p ?o .
      optional {
        ?o ?pp ?oo .
      }
    }
        Where <some_uri> is the resource we are looking at.
    """
    data = {}
    seconds = {}
    for row in results.get("results", {}).get("bindings", []):
        o_object = HashableResult(row.get("o", {}))
        data.setdefault(row.get("p", {}).get("value"), set()).add(o_object)
        if row.get("pp"):
            seconds.setdefault(o_object, {}).setdefault(
                row.get("pp", {}).get("value"), set()
            ).add(HashableResult(row.get("oo", {})))

    # now make the values into lists for easier handling
    for k, v in data.items():
        data[k] = list(sorted(v))
    for k, v in seconds.items():
        for kk, vv in v.items():
            seconds[k][kk] = list(sorted(vv))

    buf = []
    ba = buf.append
    skip_fields = []
    for title_field in (
        "https://schema.org/name",
        "http://www.w3.org/2000/01/rdf-schema#label",
        "http://www.w3.org/2004/02/skos/core#prefLabel",
    ):
        if title_field in data:
            title = list(data[title_field])[0]
            ba(H1(title, title=title_field))
            skip_fields.append(title_field)
            break

    rdf_type = data.get("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")
    if rdf_type:
        skip_fields.append("http://www.w3.org/1999/02/22-rdf-syntax-ns#type")
        rdf_types = [
            A(
                do_prefixes(str(x)),
                href=f"{MOUNT}shmarql/?query=" + make_spo(x, "o"),
                style="margin-right: 0.5em",
            )
            for x in rdf_type
        ]

        ba(P("a ", *rdf_types, style="font-style: italic;"))

    for description_field in (
        "https://schema.org/description",
        "http://schema.org/description",
    ):
        sdo_description = data.get(description_field)
        if sdo_description:
            ba(P(sdo_description[0]["value"]))
            skip_fields.append(description_field)

    for image_field in (
        "https://schema.org/logo",
        "https://schema.org/image",
        "http://schema.org/logo",
        "http://schema.org/image",
    ):
        an_image = data.get(image_field)
        if an_image:
            for i in an_image:
                ba(Img(src=i["value"], style="max-width: 100%;"))

    for field, val in data.items():
        if field in skip_fields:
            continue
        v_label_list = []
        for v in val[:50]:

            if v in seconds:
                v_label = seconds[v].get(
                    "http://www.w3.org/2000/01/rdf-schema#label", [str(v)]
                )
                v_type = seconds[v].get(
                    "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
                )
                if v["type"] == "uri":
                    v_label_list.append(
                        A(
                            v_label[0],
                            href=f"{MOUNT}shmarql/?query="
                            + make_resource_query(v["value"]),
                            style="margin-right: 1em",
                        )
                    )
                else:
                    if v_type:
                        v_label_list.append(
                            Span(v_label[0], title=v_type[0], style="margin-right: 1em")
                        )
                    else:
                        v_label_list.append(Span(v_label[0], style="margin-right: 1em"))
            else:
                v_label_list.append(
                    A(
                        v["value"],
                        href=f"{MOUNT}shmarql/?query=" + make_spo(v["value"], "o"),
                    )
                )
        P_query = f"{MOUNT}shmarql/?query=" + make_spo(field, "p")
        if len(val) > 50:
            plain_query = query.replace("# shmarql-view: resource\n", "")

            field_heading = H3(
                A(do_prefixes(field), href=P_query),
                Span(
                    f"There are {len(val)} values for this field, showing the first 50,",
                    style="font-size: 80%; font-style: italic",
                ),
                A(
                    "click here to show all",
                    href=f"{MOUNT}shmarql/?query=" + quote(plain_query),
                    style="font-size: 80%",
                ),
                style="margin: 0.5em 0 0 0",
            )
        else:
            field_heading = H3(
                A(do_prefixes(field), href=P_query), style="margin: 0.5em 0 0 0"
            )
        ba(
            (
                field_heading,
                P(
                    *[vv for vv in v_label_list],
                    style="font-size: 120%; margin: 0",
                ),
            )
        )
    return tuple(buf)


def build_standalone_table(results, query):
    table_rows = []
    heads = [Th(" ")]
    heads.extend(
        [
            Th(var, style="font-weight: bold")
            for var in results.get("head", {}).get("vars", [])
        ]
    )
    table_rows.append(Tr(*heads))

    rownum = 0
    for row in results.get("results", {}).get("bindings", []):
        rownum += 1
        row_columns = []
        row_columns.append(
            Td(
                rownum,
                style="padding-right: 0.75ch; color: #aaa; text-align: right;",
            )
        )
        for var in results.get("head", {}).get("vars", []):
            value = row.get(var, {"value": ""})
            if value.get("type") == "uri":
                S_query = make_spo(value["value"], "s")
                P_query = make_spo(value["value"], "p")
                O_query = make_spo(value["value"], "o")

                # do some fizzy for factgrid
                if value["value"].find("database.factgrid.de") > -1:
                    fizzy_query = quote(
                        f"""select distinct ?s (STR(?o) AS ?oLabel) where {{
  ?s fizzy:rdf2vec <{value['value']}> . 
  ?s rdfs:label ?o .
}}
    """
                    )
                    fizzyquery = A(
                        "✨",
                        href="?query=" + fizzy_query,
                        title="Show items similar to this entity using fizzysearch",
                    )
                else:
                    fizzyquery = None

                row_columns.append(
                    Td(
                        A(
                            "S",
                            href=f"{MOUNT}shmarql/?query={S_query}",
                            style="font-size: 80%; background-color: #ddd; color: #000; padding: 3px; text-decoration: none; margin: 0",
                        ),
                        A(
                            "P",
                            href=f"{MOUNT}shmarql/?query={P_query}",
                            style="font-size: 80%; background-color: #ddd; color: #000; padding: 3px; text-decoration: none; margin: 0",
                        ),
                        A(
                            "O",
                            href=f"{MOUNT}shmarql/?query={O_query}",
                            style="font-size: 80%; background-color: #ddd; color: #000; padding: 3px; text-decoration: none; margin: 0",
                        ),
                        A(
                            value["value"],
                            href=value["value"],
                            style="margin-left: 1ch",
                        ),
                        fizzyquery,
                        cls="border border-gray-300 px-4 py-2 text-sm",
                    )
                )
            elif value.get("type") == "bnode":
                row_columns.append(
                    Td(
                        Span(
                            value["value"], style="font-size: 80%; font-style: italic"
                        ),
                        cls="border border-gray-300 px-4 py-2 text-sm",
                    )
                )
            else:
                o_link = A(
                    "O",
                    href=f"{MOUNT}shmarql/?query={make_literal_query(value)}",
                    style="font-size: 80%; background-color: #ddd; color: #000; padding: 3px; text-decoration: none; margin: 0",
                )
                lang = (
                    Span(
                        value.get("xml:lang"), cls="text-xs bg-gray-200 text-black px-2"
                    )
                    if "xml:lang" in value
                    else None
                )

                row_columns.append(
                    Td(
                        o_link,
                        Span(value["value"], style="margin-left: 1ch"),
                        lang,
                        cls="border border-gray-300 px-4 py-2 text-sm",
                    )
                )
        table_rows.append(Tr(*row_columns, cls="hover:bg-gray-50"))
    cached = " (from cache) " if results.get("cached") else ""

    duration_display = (
        f"{int(results.get('duration', 0) * 1000)}ms"
        if results.get("duration", 0) < 1
        else f"{results.get('duration', 0):.3f}s"
    )

    return Div(
        Div(
            Span(
                f"{len(results.get('results', {}).get('bindings', []))} results in {duration_display}{cached}",
                title="used: " + results.get("endpoint_name", ""),
            ),
            A(
                "CSV",
                title="Download as CSV",
                href=f"{MOUNT}shmarql/?query={quote(query)}&format=csv",
                cls=BTN_STYLE,
            ),
            A(
                "JSON",
                title="Download as JSON",
                href=f"{MOUNT}shmarql/?query={quote(query)}&format=json",
                cls=BTN_STYLE,
            ),
            cls="bg-slate-200 text-black p-2 flex flex-row gap-1 text-xs",
        ),
        Table(
            *table_rows,
            cls="min-w-full table-auto border-collapse border border-gray-300",
        ),
    )


@rt.get(f"{MOUNT}shmarql/fragments/chart")
def fragments_chart(query: str, results=None):
    if results is None:
        results = do_query(query)

    settings = results.get("shmarql_settings", {})
    chart_type = settings.get("view")[0]

    df = results_to_df(results)

    chart_id = f"uniq-{uuid4()}"
    chart_func = {
        "barchart": do_barchart,
        "piechart": do_piechart,
        "mapchart": do_mapchart,
    }.get(chart_type)
    try:
        chart_json = to_json(chart_func(settings, df, settings.get("label", [None])[0]))
        js_options = {}
    except Exception as e:
        return Div(
            f"Chart Error: {traceback.format_exc()}",
            style="border: 2px solid red; font-size: 150%; padding: 1em;",
        )

    return Div(
        Script(
            f"""
        var plotly_data = {chart_json};
        Plotly.newPlot('{chart_id}', plotly_data.data, plotly_data.layout, {json.dumps(js_options)});
    """
        ),
        id=chart_id,
    )


def build_sparql_ui(query, results):
    settings = results.get("shmarql_settings", {})
    if "hide" in settings.get("editor", []):
        sparql_editor_block_style_button = (
            Button(
                "SPARQL",
                Script(
                    """me().on("click", async ev => {me(ev).fadeOut(); me('#sparql_editor_block').style.display = 'block'})"""
                ),
                style="padding: 4px 8px; font-size: 12px;",
                cls="md-button md-button--primary",
            ),
            Script(
                "setTimeout(() => {me('#sparql_editor_block').style.display = 'none'}, 500 )"
            ),
        )
    else:
        sparql_editor_block_style_button = None

    if "resource" in settings.get("view", []):
        results_fragment = Div(fragments_resource(results, query), cls="md-typeset")
    if settings.get("view", [""])[0].endswith("chart"):
        try:
            chart_fragment = fragments_chart(query)
        except Exception as e:
            chart_fragment = Div(
                f"Error: {e}",
                style="color: red; font-weight: bold; font-size: 150%;",
            )
        results_fragment = Div(chart_fragment, cls="md-typeset")
    else:
        results_fragment = Div(fragments_sparql(query, results), cls="md-typeset")

    return (
        Script(src=f"{MOUNT}static/editor.js"),
        Script(src=f"{MOUNT}static/matchbrackets.js"),
        Script(src=f"{MOUNT}static/sparql.js"),
        sparql_editor_block_style_button,
        Div(
            Div(
                Button(
                    "▶",
                    Script(
                        'me().on("click", async ev => {\r\n  if (ev.shiftKey) {\r\n    console.log("Shift key was pressed during the click!");\r\n  }\r\n  executeQuery()\r\n})'
                    ),
                    id="execute_sparql",
                    title="Execute this query, (also use Ctrl+Enter)",
                    style="padding: 4px 8px; font-size: 12px;",
                    cls=(ButtonT.primary),
                ),
                Button(
                    "Prefixes",
                    Script(
                        'me().on("click", async ev => {\n\n    let editorContent = sparqleditor.doc.getValue();\r\n    let prefixContent = `'
                        + PREFIXES_SNIPPET
                        + "\n\n`;\r\nsparqleditor.doc.setValue(prefixContent + editorContent);\r\n\r\n})"
                    ),
                    id="prefixes",
                    name="prefixes",
                    style="padding: 4px 8px; font-size: 12px;",
                    cls=(ButtonT.secondary, "ml-4"),
                ),
                cls="mb-2",
            ),
            Div(Textarea(query, id="code", name="code")),
            id="sparql_editor_block",
        ),
        Div(
            results_fragment,
            id="results",
            style="margin-top: 2vh;",  # max-height: 50vh; overflow-y: scroll
        ),
        Script(
            f"""
const link = document.createElement("link");
  link.rel = "stylesheet";
  link.href = "{MOUNT}static/codemirror.css";
  document.head.appendChild(link);"""
        ),
        Script(f"const MOUNT = '{MOUNT}';"),
        Script(
            """
document.addEventListener("DOMContentLoaded", function () {
  sparqleditor = CodeMirror.fromTextArea(document.getElementById("code"), {
    mode: "application/sparql-query",
    matchBrackets: true,
    lineNumbers: true,
  });
  results = document.getElementById("results");
});

function executeQuery() {
    let the_query = sparqleditor.doc.getValue();
    results.innerHTML = '<div aria-busy="true">Loading...</div>';
    history.pushState({ query: the_query }, "", MOUNT+"shmarql/?query=" + encodeURIComponent(the_query));
    queryStarted = Date.now();
    progress_counter = setTimeout(updateProgress, 1000);
    
    fetch(`${MOUNT}shmarql/fragments/sparql`, {
        method: "POST",
        headers: {
            "Content-Type": "application/x-www-form-urlencoded"
        },
        body: `query=${encodeURIComponent(the_query)}`
    }).then(response => response.text()).then(data => {
        if(progress_counter) {
            clearTimeout(progress_counter);
        }
        results.innerHTML = data;
        const scripts = results.querySelectorAll("script");
            scripts.forEach((script) => {
                const newScript = document.createElement("script");
                if (script.src) {
                    newScript.src = script.src;
                    newScript.async = false; // To preserve execution order if needed
                } else {
                    newScript.textContent = script.textContent;
                }
                document.body.appendChild(newScript);
                document.body.removeChild(newScript);
        });
    })

}

function updateProgress() {
    let progress = Math.round((Date.now() - queryStarted) / 1000);
    results.innerHTML = `<div aria-busy="true">Query in progress, took ${progress}s so far...</div>`;
    progress_counter = setTimeout(updateProgress, 1000);
}

document.body.addEventListener("keypress", function (evt) {
    if (evt.ctrlKey && evt.key === "Enter") {
        evt.preventDefault();
        executeQuery()
    }
});
"""
        ),
    )



from fasthtml.common import *
import json
import bikidata
from .main import app
from .config import MOUNT


def results_to_div(r: dict):
    if "error" in r:
        return Div(
            H3("Error"),
            P(r["error"]),
        )

    hits = r.get("results", {})

    agg_buf = []
    for prop, aggs in r.get("aggregates", {}).items():
        agg_buf.append(H3(f"Aggregates: {prop}"))
        agg_buf.append(Table(*[Tr(Td(iri), Td(str(c))) for c, iri in aggs]))

    extra_text = "."
    if r["total"] > r["size"]:
        extra_text = f", showing the first {r['size']}"

    return Div(
        Div(
            *agg_buf,
            style="margin-bottom: 3ch",
        ),
        P(f"{r['total']} results in total{extra_text}"),
        *[
            Div(
                A(hitiri, target="_whatsnew", href=hitiri.strip("<>")),
                P(
                    fields.get(
                        "<http://www.w3.org/2000/01/rdf-schema#label>",
                        [""],
                    )[
                        0
                    ][:300]
                    .strip('"')
                    .replace('"@en', "")
                    .replace('"@de', "")
                    .replace(r"\"", '"'),
                    style="padding: 0.2ch;",
                ),
                style="border: 1px solid #ccc; padding: 1ch",
            )
            for hitiri, fields in hits.items()
        ],
    )


@app.post(MOUNT + "bikidata")
async def query(request: Request):
    body = await request.body()
    opts = json.loads(body)
    try:
        r = bikidata.query(opts)
    except Exception as e:
        r = {"error": str(e)}

    if opts.get("format") == "json":
        return JSONResponse(r)

    return results_to_div(r)


@app.get(MOUNT + "bikidata/")
def biki_get():
    return Div(
        Textarea(
            """{
    "filters": [
        {"p":"fts 2", "o":"culture"}
        ],
    "aggregates": [
         "<http://www.w3.org/1999/02/22-rdf-syntax-ns#type>"
    ],
    "size": 20
}""",
            id="opts",
            style="width: 80%; height: 20ch",
        ),
        Script(
            """me().on("keyup", async event => {
if(event.code === "Enter" && event.ctrlKey) {
    fetch('"""
            + MOUNT
            + """bikidata', {
        method: "POST",
        body: me("#opts").value,
    }).then(r => r.text()).then(r => me("#results").innerHTML = r);
}

})
"""
        ),
        Div(id="results"),
    )


@app.get(MOUNT + "bikidata/browse")
def biki_browse():
    clipboard = """<svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" fill="currentColor" class="bi bi-clipboard" viewBox="0 0 24 24">
    <path d="M4 1.5H3a2 2 0 0 0-2 2V14a2 2 0 0 0 2 2h10a2 2 0 0 0 2-2V3.5a2 2 0 0 0-2-2h-1v1h1a1 1 0 0 1 1 1V14a1 1 0 0 1-1 1H3a1 1 0 0 1-1-1V3.5a1 1 0 0 1 1-1h1v-1z"/>
    <path d="M9.5 1a.5.5 0 0 1 .5.5v1a.5.5 0 0 1-.5.5h-3a.5.5 0 0 1-.5-.5v-1a.5.5 0 0 1 .5-.5h3zm-3-1A1.5 1.5 0 0 0 5 1.5v1A1.5 1.5 0 0 0 6.5 4h3A1.5 1.5 0 0 0 11 2.5v-1A1.5 1.5 0 0 0 9.5 0h-3z"/>
</svg>"""

    props = bikidata.query({"aggregates": ["properties"]})
    props = sorted(
        [
            (count, piri)
            for count, piri in props.get("aggregates", {}).get("properties", [])
        ],
        reverse=True,
    )

    op = Div(
        Select(
            Option("must", value="must"),
            Option("should", value="should"),
            Option("not", value="not"),
            cls="op",
            style="margin: 1ch; padding: 0.5ch;",
        ),
        style="display: inline-block; margin: 0.1ch;",
    )
    predicate = Div(
        Select(
            Option("fts", value="fts"),
            Option(" - ", value="_", title="No value"),
            Option("id", value="id"),
            Option("semantic", value="semantic"),
            *[Option(f"{c} {p}", value=p) for c, p in props],
            cls="pred",
            style="margin: 1ch; padding: 0.5ch;",
        ),
        style="display: inline-block; margin: 0.1ch;",
    )
    obj = Div(
        Input(type="text", style="width: 30vw", cls="obj"),
        style="display: inline-block; margin: 0.1ch;",
    )

    block = Div(
        Button(
            "-",
            style="padding: 0.2ch; border: 1px solid #ccc; border-radius: 8px;",
            cls="block_remove",
        ),
        Button(
            "+",
            style="padding: 0.2ch; border: 1px solid #ccc; border-radius: 8px;",
            cls="block_add",
        ),
        op,
        predicate,
        obj,
        style="border-bottom: 1px solid #bbb",
    )

    aggregates_chooser = Div(
        Button(
            "-",
            style="padding: 0.2ch; border: 1px solid #ccc; border-radius: 8px;",
            cls="agg_remove",
        ),
        Button(
            "+",
            style="padding: 0.2ch; border: 1px solid #ccc; border-radius: 8px;",
            cls="agg_add",
        ),
        Select(
            Option(" - ", value="-"),
            *[Option(p, value=p) for c, p in props],
            cls="agg",
            style="margin: 1ch; padding: 0.5ch;",
        ),
        style="margin: 0.1ch;",
    )

    return (
        (
            Script(type="module", src="https://pyscript.net/releases/2025.3.1/core.js"),
            Div(
                H4("Filters"),
                block,
                H4("Aggregates"),
                aggregates_chooser,
                Div(
                    Button(
                        "Go",
                        id="go",
                        style="background-color: #999; color: white; border-radius: 8px; padding: 4px",
                    ),
                    Div(id="results"),
                ),
            ),
        ),
        Script(
            """from pyscript import display, when, fetch, window
from pyscript.web import page
import json

@when("click", ".block_add")
def block_add_click(event):
    block = event.target.parentElement
    new_block = block.cloneNode(True)
    new_block.querySelector(".block_remove").onclick = block_remove_click
    new_block.querySelector(".block_add").onclick = block_add_click
    block.parentElement.insertBefore(new_block, block.nextSibling)

@when("click", ".block_remove")
def block_remove_click(event):
    existing = page.find(".block_remove")
    if len(existing) == 1:
        return
    block = event.target.parentElement
    block.parentElement.removeChild(block)

@when("click", ".agg_add")
def agg_add_click(event):
    block = event.target.parentElement
    new_block = block.cloneNode(True)
    new_block.querySelector(".agg_remove").onclick = agg_remove_click
    new_block.querySelector(".agg_add").onclick = agg_add_click
    block.parentElement.insertBefore(new_block, block.nextSibling)    


@when("click", ".agg_remove")
def agg_remove_click(event):
    existing = page.find(".agg_remove")
    if len(existing) == 1:
        return
    block = event.target.parentElement
    block.parentElement.removeChild(block)    


@when("click", "#go")
async def go_click(event):
    # await window.navigator.clipboard.writeText("een twee drie")

    ops = page.find(".op")
    preds = page.find(".pred")
    objs = page.find(".obj")

    opts = {}
    for i, op in enumerate(ops):
        opts[i] = {'op': op.value}
    for i, pred in enumerate(preds):
        if pred.value != "_":
            opts[i]['p'] = pred.value
    for i, obj in enumerate(objs):
        if obj.value:
            opts[i]['o'] = obj.value
    opts = {'filters': [x for x in opts.values()]}

    aggs = page.find(".agg")
    if len(aggs) > 0:
        opts['aggregates'] = [agg.value for agg in aggs if agg.value != "-"]

    opts['format'] = 'html'    

    response = await fetch('/bikidata', 
        method="POST",
        headers={ "Content-Type": "application/json"},
        body =  json.dumps(opts)
    )
    if response.ok:
        data = await response.text()        
        page.find("#results").innerHTML = data
    else:
        print(response.status)
""",
            type="mpy",
        ),
    )



import os, sqlite3, random, json, logging

log = logging.getLogger("SHMARQL")
handler = logging.StreamHandler()
log.addHandler(handler)


DEBUG = os.environ.get("DEBUG", "0") == "1"
if DEBUG:
    log.setLevel(logging.DEBUG)
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(levelname)-9s %(name)s %(asctime)s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    handler.setFormatter(formatter)
    log.debug("Debug logging requested from config env DEBUG")
else:
    log.setLevel(logging.INFO)
    log.info("SHMARQL Logging at INFO level")


ENDPOINT = os.environ.get("ENDPOINT")

# ENDPOINTS variable with name|url pairs
ens = os.environ.get("ENDPOINTS", "")

# Split the string into name|url pairs and then further split each pair
ens_pairs = [pair.split("|") for pair in ens.split(" ") if "|" in pair]

# Convert into a dictionary
ENDPOINTS = {name: url for name, url in ens_pairs}

SCHEME = os.environ.get("SCHEME", "http://")
DOMAIN = os.environ.get("DOMAIN", "127.0.0.1")
PORT = os.environ.get("PORT", "5001")
# Note, we can't just build a SITE_URI from the above variables because the
# app might be running behind a reverse proxy
SITE_URI = os.environ.get("SITE_URI", "http://127.0.0.1:8000/")

# This is a mountpoint that will be prefixed to all URIs served by the application
MOUNT = os.environ.get("MOUNT", "/")

QUERIES_DB = os.environ.get("QUERIES_DB", "queries.db")
thequerydb = sqlite3.connect(QUERIES_DB)
thequerydb.executescript(
    """CREATE TABLE IF NOT EXISTS queries (queryhash TEXT, query TEXT, timestamp TEXT, endpoint TEXT, result TEXT, duration FLOAT);
pragma journal_mode=WAL;"""
)

if "DATA_LOAD_PATHS" in os.environ:
    DATA_LOAD_PATHS = os.environ.get("DATA_LOAD_PATHS").split(" ")
else:
    DATA_LOAD_PATHS = []
STORE_PATH = os.environ.get("STORE_PATH")

BIKIDATA_DB = os.environ.get("BIKIDATA_DB")
SEMANTIC_INDEX = os.environ.get("SEMANTIC_INDEX", "0") == "1"
RDF2VEC_INDEX = os.environ.get("RDF2VEC_INDEX", "0") == "1"

FTS_FILEPATH = os.environ.get("FTS_FILEPATH")
RDF2VEC_FILEPATH = os.environ.get("RDF2VEC_FILEPATH")

if FTS_FILEPATH:
    log.exception(
        "FTS_FILEPATH is set, but this config has been removed. Please use BIKIDATA_DB instead. See https://shmarql.com/fizzysearch/ for more information."
    )
    FTS_FILEPATH = None
if RDF2VEC_FILEPATH:
    log.exception(
        "RDF2VEC_FILEPATH is set, but this config has been removed. Please use RDF2VEC_INDEX boolean and BIKIDATA_DB instead. See https://shmarql.com/rdf2vec/ for more information."
    )
    RDF2VEC_FILEPATH = None


SPARQL_QUERY_UI = os.environ.get("SPARQL_QUERY_UI", "1") == "1"

SITE_ID = os.environ.get(
    "SITE_ID", "".join([random.choice("abcdef0123456789") for _ in range(10)])
)

SITEDOCS_PATH = os.environ.get("SITEDOCS_PATH", os.path.join(os.getcwd(), "site"))
SCHPIEL_PATH = os.environ.get("SCHPIEL_PATH")

PREFIXES_FILEPATH = os.environ.get("PREFIXES_FILEPATH")
DEFAULT_PREFIXES = {
    "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf:",
    "http://www.w3.org/2000/01/rdf-schema#": "rdfs:",
    "http://www.w3.org/2002/07/owl#": "owl:",
    "http://schema.org/": "schema:",
    "http://www.wikidata.org/entity/": "wd:",
    "http://www.wikidata.org/entity/statement/": "wds:",
    "http://wikiba.se/ontology#": "wikibase:",
    "http://www.wikidata.org/prop/direct/": "wdt:",
    "http://www.w3.org/2004/02/skos/core#": "skos:",
    "http://purl.org/dc/terms/": "dct:",
    "http://purl.org/dc/elements/1.1/": "dc:",
    "http://dbpedia.org/resource/": "dbr:",
    "https://www.ica.org/standards/RiC/ontology#": "rico:",
    "http://www.w3.org/2003/01/geo/wgs84_pos#": "geo:",
    "http://www.w3.org/ns/shacl#": "sh:",
    "http://www.w3.org/2001/XMLSchema#": "xsd:",
    "http://www.openlinksw.com/virtrdf-data-formats#": "virtrdfdata:",
    "http://www.openlinksw.com/schemas/virtrdf#": "virtrdf:",
    "http://purl.org/spar/fabio/": "fabio:",
    "http://www.w3.org/2003/11/swrl#": "swrl:",
    "http://www.w3.org/ns/dcat#": "dcat:",
    "https://shmarql.com/": "shmarql:",
    "https://database.factgrid.de/entity/": "factgrid:",
}

try:
    if PREFIXES_FILEPATH:
        # also support reading the prefixed from a .ttl file for convenience
        if PREFIXES_FILEPATH.endswith(".ttl"):
            PREFIXES = DEFAULT_PREFIXES
            for line in open(PREFIXES_FILEPATH).readlines():
                if not line.lower().startswith("@prefix "):
                    continue
                if not line.lower().endswith(" .\n"):
                    continue
                line = line.strip("\n .")
                parts = line.split(":")
                if len(parts) < 2:
                    continue
                prefix = parts[0][8:] + ":"
                prefix_uri = ":".join(parts[1:]).strip("<> ")
                if prefix == ":":
                    prefix = " "
                PREFIXES[prefix_uri] = prefix
        else:
            PREFIXES = DEFAULT_PREFIXES | json.load(open(PREFIXES_FILEPATH))
    else:
        PREFIXES = DEFAULT_PREFIXES
except:
    log.exception(f"Problem binding PREFIXES from {PREFIXES_FILEPATH}")

PREFIXES_SNIPPET = "".join(
    f"PREFIX {prefix} <{uri}>\n" for uri, prefix in PREFIXES.items()
)

ADMIN_DATABASE = os.environ.get("ADMIN_DATABASE")

LOGINS = os.environ.get("LOGINS") == "1"

SITE_TITLE = os.environ.get("SITE_TITLE", "SHMARQL")



from fasthtml.common import *
from monsterui.all import *
from mistletoe.contrib.pygments_renderer import PygmentsRenderer
from urllib.parse import quote
from .fragments import fragments_sparql
from .config import LOGINS, SITE_TITLE, ENDPOINT, DATA_LOAD_PATHS
import yaml


def navbar(session={}):
    user = session.get("user")
    if LOGINS:
        login_link = (
            A(
                f"Logout {user.get('username')}",
                href="/logout",
                cls="btn btn-sm btn-outline",
            )
            if user
            else A("Login", href="/login", cls="btn btn-sm btn-primary")
        )

    return Div(
        NavBar(
            A("Query", href="/shmarql/") if ENDPOINT or DATA_LOAD_PATHS else None,
            (
                A("Users", href="/admin/users", cls="btn btn-sm btn-outline")
                if user and user.get("username") == "admin"
                else None
            ),
            login_link if LOGINS else None,
            brand=DivLAligned(
                A(UkIcon("home", height=30, width=30), href="/"),
                A(H3(SITE_TITLE, cls="text-zinc-100"), href="/"),
            ),
        ),
        cls="bg-sky-600 text-white",
    )


def footer():
    return Div(
        "Made with ❤️ in Amsterdam by ",
        A("epoz", href="https://epoz.org", target="_blank", cls="text-sky-200"),
        " for the ",
        A(
            "ISE Group at FIZ Karlsruhe",
            href="https://www.fiz-karlsruhe.de/en/bereiche/information-service-engineering",
            target="_blank",
        ),
        cls="bg-slate-800 text-center text-zinc-500 pt-4 pb-4 mt-8",
    )


def base(content, extra_script=None, title="SHMARQL", session={}):
    return (
        Title(title),
        extra_script,
        navbar(session=session),
        Container(content, cls="lg:w-4/5 w-11/12"),
        footer(),
    )


class LanguageAwarePygmentsRenderer(PygmentsRenderer):
    def render_block_code(self, token):
        tmp = super().render_block_code(token)

        query = quote(token.content)
        runlink = to_xml(
            A(
                UkIcon("send", height=30, width=30),
                href=f"/shmarql/?query={query}",
            )
        )

        if token.language == "shmarql":
            query_result = to_xml(fragments_sparql(token.content))
            return f"{runlink}{query_result}"
        elif token.language == "sparql":
            return f'{runlink}<div class="code-block" data-language="{token.language}">\n{tmp}\n</div>'
        else:
            return f'<div class="code-block" data-language="{token.language}">\n{tmp}\n</div>'


def build_nav(navlist: list, depth=0.25):
    buf = []
    for navitem in navlist:
        for title, item in navitem.items():
            if isinstance(item, str):
                buf.append(Li(A(title, href=item), cls="mt-2"))
            else:
                buf.append(P(title, cls=TextT.bold))
                buf.append(build_nav(item, depth=depth + 0.25))
    return Ul(*buf, cls=(f"pr-2", TextT.sm), style=f"padding-left: {depth}ch")


def markdown_container(filepath: str, nav=[], session={}):
    sidebar = None
    if nav:
        sidebar = Div(build_nav(nav), cls="pt-11 w-3/5")

    md_content = open(filepath, "r").read()
    c = render_md(md_content, renderer=LanguageAwarePygmentsRenderer)
    return base(Div(sidebar, Div(c), cls="flex gap-12"), session=session)



import click
import yaml
from shmarql.config import log
from mkdocs.__main__ import cli as mkdocs_cli
from .am import reset_admin_password


@click.option(
    "-f",
    "--filepath",
    type=click.Path(exists=True),
    help="Path to the admin database file",
    required=True,
)
@click.command("reset_admin")
def reset_admin(filepath: str):
    reset_admin_password(filepath)


@click.option(
    "-f",
    "--filepath",
    type=click.Path(exists=True),
    help="Path to the mkdocs.yml file",
)
@click.command("docs_build")
def docs_build(filepath: str):
    log.debug(f"Trying to open filepath:   {filepath}")
    nav = yaml.safe_load(open(filepath).read())
    SRC_MKDOCS = "mkdocs.yml"
    log.debug(f"Assuming the site mkdocs.yml file is in:   {SRC_MKDOCS}")
    try:
        site_mkdocs = yaml.load(open(SRC_MKDOCS).read(), yaml.UnsafeLoader)
    except FileNotFoundError:
        log.error(f"No {SRC_MKDOCS} file found, exiting.")
        return
    changed = False
    for key in ("site_name", "site_url", "repo_url", "nav"):
        if key in site_mkdocs:
            del site_mkdocs[key]
    for key in ("site_name", "site_url", "repo_url", "nav", "plugins"):
        if key in nav:
            site_mkdocs[key] = nav[key]
            changed = True

    open(SRC_MKDOCS, "w").write(yaml.dump(site_mkdocs))
    click.echo(f"Wrote new {SRC_MKDOCS} file")

    try:
        mkdocs_cli(["build", "--site-dir", "site"], standalone_mode=False)
    except Exception as e:
        log.error(str(e))


@click.group()
def cli():
    pass


cli.add_command(docs_build)
cli.add_command(reset_admin)

if __name__ == "__main__":
    cli()



from pyoxigraph import (
    NamedNode,
    Literal,
    BlankNode,
    Quad,
    QuerySolutions,
    QueryTriples,
    Store,
    Variable,
    RdfFormat,
)
from io import BytesIO
import logging
from rdflib.plugins.sparql.results.xmlresults import SPARQLXMLWriter
from xml.sax.xmlreader import AttributesNSImpl
from xml.dom import XML_NAMESPACE
import pandas as pd
from typing import Union
from .config import PREFIXES


def string_iterator(graph: Store):
    for s, p, o, g in graph.quads_for_pattern(None, None, None, None):
        yield (str(s), str(p), str(o), str(g))


class SerializationException(Exception): ...


def termJSON(term):
    if isinstance(term, NamedNode):
        return {"type": "uri", "value": term.value}
    elif isinstance(term, Literal):
        r = {"type": "literal", "value": term.value}
        if (
            term.datatype is not None
            and term.datatype.value != "http://www.w3.org/2001/XMLSchema#string"
        ):
            r["datatype"] = term.datatype.value
        if term.language is not None:
            r["xml:lang"] = term.language
        return r
    elif isinstance(term, BlankNode):
        return {"type": "bnode", "value": term.value}
    elif term is None:
        return None
    else:
        raise SerializationException("Unknown term type: %s (%s)" % (term, type(term)))


class OxigraphSerialization:
    def __init__(self, result):
        self.result = result

    def json(self):
        if type(self.result) in (QuerySolutions, SynthQuerySolutions):
            return self.qr_json()
        elif type(self.result) == QueryTriples:
            return self.qt_json()

    def to_store(self) -> Store:
        tmp_store = Store()
        tmp_store.extend([Quad(s, p, o) for s, p, o in self.result])
        return tmp_store

    def qt_json(self):

        result = {
            "data": self.result.serialize(format=RdfFormat.N_TRIPLES).decode("utf8")
        }
        return result

    def qt_turtle(self):
        tmp_store = self.to_store()
        buf = BytesIO()
        tmp_store.dump(buf, "text/turtle")
        return buf.getvalue().decode("utf8")

    def qr_json(self):
        result = {"head": {"vars": [x.value for x in self.result.variables]}}
        rows = []
        for qs in self.result:
            row = {}
            for var in self.result.variables:
                if qs[var] is not None:
                    row[var.value] = termJSON(qs[var])
            rows.append(row)
        result["results"] = {"bindings": rows}
        return result

    def xml(self):
        SPARQL_XML_NAMESPACE = "http://www.w3.org/2005/sparql-results#"
        if type(self.result) in (QuerySolutions, SynthQuerySolutions):
            stream = BytesIO()
            sxw = SPARQLXMLWriter(stream, encoding="utf-8")
            vars = [v.value for v in self.result.variables]
            sxw.write_header(vars)
            sxw.write_results_header()
            for qs in self.result:
                sxw.write_start_result()
                for name in vars:
                    if qs[name] is not None:
                        attr_vals = {
                            (None, "name"): str(name),
                        }
                        attr_qnames = {
                            (None, "name"): "name",
                        }
                        sxw.writer.startElementNS(
                            (SPARQL_XML_NAMESPACE, "binding"),
                            "binding",
                            AttributesNSImpl(attr_vals, attr_qnames),
                        )
                        val = qs[name]
                        if type(val) == NamedNode:
                            sxw.writer.startElementNS(
                                (SPARQL_XML_NAMESPACE, "uri"),
                                "uri",
                                AttributesNSImpl({}, {}),
                            )
                            sxw.writer.characters(val.value)
                            sxw.writer.endElementNS(
                                (SPARQL_XML_NAMESPACE, "uri"), "uri"
                            )
                        elif type(val) == BlankNode:
                            sxw.writer.startElementNS(
                                (SPARQL_XML_NAMESPACE, "bnode"),
                                "bnode",
                                AttributesNSImpl({}, {}),
                            )
                            sxw.writer.characters(val.value)
                            sxw.writer.endElementNS(
                                (SPARQL_XML_NAMESPACE, "bnode"), "bnode"
                            )
                        elif type(val) == Literal:
                            attr_vals = {}
                            attr_qnames = {}
                            if val.language:
                                attr_vals[(XML_NAMESPACE, "lang")] = val.language
                                attr_qnames[(XML_NAMESPACE, "lang")] = "xml:lang"
                            elif val.datatype:
                                attr_vals[(None, "datatype")] = val.datatype.value
                                attr_qnames[(None, "datatype")] = "datatype"

                            sxw.writer.startElementNS(
                                (SPARQL_XML_NAMESPACE, "literal"),
                                "literal",
                                AttributesNSImpl(attr_vals, attr_qnames),
                            )
                            sxw.writer.characters(val.value)
                            sxw.writer.endElementNS(
                                (SPARQL_XML_NAMESPACE, "literal"), "literal"
                            )

                        sxw.writer.endElementNS(
                            (SPARQL_XML_NAMESPACE, "binding"), "binding"
                        )

                sxw.write_end_result()
            sxw.close()
            return stream.getvalue().decode("utf8")


class SynthQuerySolutions:
    variables = [Variable("s"), Variable("p"), Variable("o")]

    def __init__(self, triples):
        self.triples = [
            {
                Variable("s"): s,
                Variable("p"): p,
                Variable("o"): o,
            }
            for s, p, o in triples
        ]

    def __iter__(self):
        self.index = 0
        return self

    def __next__(self):
        if self.index >= len(self.triples):
            raise StopIteration
        current_variable = self.triples[self.index]
        self.index += 1
        return current_variable


def results_to_triples(results: dict, vars: dict):
    buf = []
    if "results" in results and "bindings" in results["results"]:
        for row in results["results"]["bindings"]:
            line = []
            for tmp in ("s", "p", "o"):
                x = row.get(tmp)
                if not x is None:
                    if x.get("type") in ("literal", "typed-literal"):
                        if tmp == "p":
                            logging.debug(f"Bogus literal as predicate: {x}")
                            line.append(
                                NamedNode(
                                    "http://bogus_literal_as_predicate_from_endpoint/"
                                )
                            )
                            continue

                        datatype = x.get("datatype")
                        if datatype:
                            line.append(
                                Literal(
                                    x.get("value"),
                                    language=x.get("xml:lang"),
                                    datatype=NamedNode(datatype),
                                )
                            )
                        else:
                            line.append(
                                Literal(x.get("value"), language=x.get("xml:lang"))
                            )
                            continue
                    elif x.get("type") == "uri":
                        line.append(NamedNode(x.get("value")))
                        continue
                    elif x.get("type") == "bnode":
                        line.append(BlankNode(x.get("value")))
                        continue
                if vars.get(tmp) != "?" + tmp:
                    varvalue = vars.get(tmp)
                    if varvalue.startswith("<"):
                        line.append(NamedNode(varvalue.strip("<>")))
                        continue
                    elif varvalue.startswith("_"):
                        line.append(BlankNode(varvalue[1:]))
                        continue
                    else:
                        line.append(Literal(varvalue))
                        continue
            buf.append(line)
    return buf


def results_to_xml(results: dict) -> str:
    SPARQL_XML_NAMESPACE = "http://www.w3.org/2005/sparql-results#"
    stream = BytesIO()
    sxw = SPARQLXMLWriter(stream, encoding="utf-8")
    vars = results.get("head", {}).get("vars", [])
    sxw.write_header(vars)
    sxw.write_results_header()
    for qs in results.get("results", {}).get("bindings", []):
        sxw.write_start_result()
        for name in vars:
            if qs.get(name) is not None:
                attr_vals = {
                    (None, "name"): name,
                }
                attr_qnames = {
                    (None, "name"): "name",
                }
                sxw.writer.startElementNS(
                    (SPARQL_XML_NAMESPACE, "binding"),
                    "binding",
                    AttributesNSImpl(attr_vals, attr_qnames),
                )
                val = qs.get(name)
                if val.get("type") == "uri":
                    sxw.writer.startElementNS(
                        (SPARQL_XML_NAMESPACE, "uri"),
                        "uri",
                        AttributesNSImpl({}, {}),
                    )
                    sxw.writer.characters(val.get("value"))
                    sxw.writer.endElementNS((SPARQL_XML_NAMESPACE, "uri"), "uri")
                elif val.get("type") == "bnode":
                    sxw.writer.startElementNS(
                        (SPARQL_XML_NAMESPACE, "bnode"),
                        "bnode",
                        AttributesNSImpl({}, {}),
                    )
                    sxw.writer.characters(val.get("value"))
                    sxw.writer.endElementNS((SPARQL_XML_NAMESPACE, "bnode"), "bnode")
                elif type(val) == "literal":
                    attr_vals = {}
                    attr_qnames = {}
                    if val.get("xml:lang"):
                        attr_vals[(XML_NAMESPACE, "lang")] = val.get("xml:lang")
                        attr_qnames[(XML_NAMESPACE, "lang")] = "xml:lang"
                    elif val.get("datatype"):
                        attr_vals[(None, "datatype")] = val.get("datatype")
                        attr_qnames[(None, "datatype")] = "datatype"

                    sxw.writer.startElementNS(
                        (SPARQL_XML_NAMESPACE, "literal"),
                        "literal",
                        AttributesNSImpl(attr_vals, attr_qnames),
                    )
                    sxw.writer.characters(val.get("value"))
                    sxw.writer.endElementNS(
                        (SPARQL_XML_NAMESPACE, "literal"), "literal"
                    )

                sxw.writer.endElementNS((SPARQL_XML_NAMESPACE, "binding"), "binding")

        sxw.write_end_result()
    sxw.close()
    return stream.getvalue().decode("utf8")


def results_to_df(results: dict) -> pd.DataFrame:
    data = {}
    vars = results.get("head", {}).get("vars", [])
    for row in results.get("results", {}).get("bindings", []):
        for var in vars:
            val = row.get(var)
            data.setdefault(var, []).append(val.get("value") if val else None)
    # this loses the type of the item in the results!
    return pd.DataFrame(data)


def do_prefixes(iris: Union[str, list]):
    """Given a list of IRI values, return a string with the IRIs prefixed"""
    if isinstance(iris, str):
        iris = [iris]
    buf = []
    for iri in iris:
        found = False
        for uri, prefix in PREFIXES.items():
            if iri.startswith(uri):
                buf.append(f"{prefix}{iri[len(uri):]}")
                found = True
        if not found:
            buf.append(iri)
    return " ".join(buf)
